/* ----------------------------------------------------------------------------
 *         ATMEL Microcontroller Software Support 
 * ----------------------------------------------------------------------------
 * Copyright (c) 2008, Atmel Corporation
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * Atmel's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
 */

//-----------------------------------------------------------------------------
/// \dir "Basic EMAC uIP project"
///
/// !!!Purpose
///
///  This project implements two examples of the
///   <a href="../../../external_libs/ethernet/uip/doc/html/index.html"> uIP TCP/IP stack</a>: 
///    - web-server.  Example enables the device to act as a web server, 
///       displaying network information through an HTML browser. 
///    - hello-world.  Example makes the device responds to telnet connections on port 1000. 
///
/// !!!Hardware requirements
///
///  The system reset signal NRST is connected to both the Ethernet PHY and the JTAG/ICE interface. 
///  This prevents the project from working properly when a JTAG probe is connected. 
///  To disconnect NRST from the ICE interface, the following modifications must be performed: 
///   - AT91SAM7X/XC-EK: cut strap S2. 
///   - AT91SAM9XE-EK: remove resistor R14 
///   - AT91SAM9260-EK: remove resistor R14 
///   - AT91SAM9263-EK: remove resistor R18 
///
/// !!!See
///
///   - emac: EMAC interface driver
///   - dm9161: PHY interface driver
///
/// !!!Description
///
/// Please refer to the <a href="../../../external_libs/ethernet/uip/doc/html/index.html"> uIP documentation</a>  for more information
/// about the TCP/IP stack, the hello-world and the webserver examples. 
///
/// If yo don't want to use DHCP, please open file uip-conf.h
/// and comment the line #define UIP_DHCP_on
///
/// !!!Usage
///
/// -# Build the program and download it inside the evaluation board. Please
///    refer to the
///    <a href="http://www.atmel.com/dyn/resources/prod_documents/doc6224.pdf">
///    SAM-BA User Guide</a>, the
///    <a href="http://www.atmel.com/dyn/resources/prod_documents/doc6310.pdf">
///    GNU-Based Software Development</a> application note or to the
///    <a href="ftp://ftp.iar.se/WWWfiles/arm/Guides/EWARM_UserGuide.ENU.pdf">
///    IAR EWARM User Guide</a>, depending on your chosen solution.
/// -# On the computer, open and configure a terminal application
///    (e.g. HyperTerminal on Microsoft Windows) with these settings:
///   - 115200 bauds
///   - 8 bits of data
///   - No parity
///   - 1 stop bit
///   - No flow control
/// -# Connect an Ethernet cable between the evaluation board and the network. 
///     The board may be connected directly to a computer; in this case, 
///     make sure to use a cross/twisted wired cable such as the one provided with the evaluation kit. 
/// -# Start the application. It will display the following message on the DBGU: 
///    \code   
///    -- Basic EMAC uIP Project xxx --
///    -- AT91xxxxxx-xx
///    -- Compiled: xxx xx xxxx xx:xx:xx --
///    - MAC 00:45:56:78:9a:bc
///    - Host IP 10.159.245.156
///    - Router IP 10.159.245.157
///    - Net Mask 255.255.254.0
///    \endcode
/// -# For the webserver example, type the IP address (Host IP in the debug log) of the device in a web browser. 
///     The page generated by lwIP will appear. 
/// -# For the hello-world example, connect to the device IP address using telnet on port 1000. A greeting message will appear. 
///
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/// \unit
///
/// !Purpose
///
/// This file contains all the specific code for the
/// basic-emac-uip-project
///
/// !Contents
///
/// The code can be roughly broken down as follows:
///    - Init EMAC driver structure
///    - Init DM9161 driver
///    - Initialize uIP modules
///    - Set the IP address of this host
///    - httpd_init or hello_world_init 
///    - uip process
///
/// Please refer to the list of functions in the #Overview# tab of this unit
/// for more detailed information.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//         Headers
//-----------------------------------------------------------------------------

#include "uip.h"
#include "uip_arp.h"
#include "tapdev.h"
#include "timer.h"

//-----------------------------------------------------------------------------
//         Local Define
//-----------------------------------------------------------------------------

/// uIP buffer : The ETH header
#define BUF ((struct uip_eth_hdr *)&uip_buf[0])

//-----------------------------------------------------------------------------
//         Global functions
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
/// Global function for uIP to use
/// \param m Pointer to string that logged
//-----------------------------------------------------------------------------
void uip_log(char *m)
{
    TRACE_INFO("-uIP- %s\n\r", m);
}

#ifdef __DHCPC_H__
//-----------------------------------------------------------------------------
/// Global function for uIP DHCPC to use, notification of DHCP configuration
/// \param s Pointer to DHCP state instance
//-----------------------------------------------------------------------------
void dhcpc_configured(const struct dhcpc_state *s)
{
    u8_t * pAddr;

    printf("\n\r");
    printf("=== DHCP Configurations ===\n\r");
    pAddr = (u8_t *)s->ipaddr;
    printf("- IP   : %d.%d.%d.%d\n\r",
        pAddr[0], pAddr[1], pAddr[2], pAddr[3]);
    pAddr = (u8_t *)s->netmask;
    printf("- Mask : %d.%d.%d.%d\n\r",
        pAddr[0], pAddr[1], pAddr[2], pAddr[3]);
    pAddr = (u8_t *)s->default_router;
    printf("- GW   : %d.%d.%d.%d\n\r",
        pAddr[0], pAddr[1], pAddr[2], pAddr[3]);
    pAddr = (u8_t *)s->dnsaddr;
    printf("- DNS  : %d.%d.%d.%d\n\r",
        pAddr[0], pAddr[1], pAddr[2], pAddr[3]);
    printf("===========================\n\r\n");
    uip_sethostaddr(s->ipaddr);
    uip_setnetmask(s->netmask);
    uip_setdraddr(s->default_router);

  #ifdef __RESOLV_H__
    resolv_conf(s->dnsaddr);
  #else
    printf("DNS NOT enabled in the demo\n\r");
  #endif
}
#endif /* __DHCPC_H__ */

//-----------------------------------------------------------------------------
/// Initialize demo application
//-----------------------------------------------------------------------------
static void app_init(void)
{
    printf("P: APP Init ... ");
    printf("hello-world\n\r");
    hello_world_init();

#ifdef __DHCPC_H__
    printf("P: DHCPC Init\n\r");
    dhcpc_init(MacAddress.addr, 6);
#endif
}

//-----------------------------------------------------------------------------
/// Default main() function.
/// Do initialization and process tasks.
//-----------------------------------------------------------------------------
int main(void)
{
    uip_ipaddr_t ipaddr;
    struct timer periodic_timer, arp_timer;
    unsigned int i;

    // System devices initialize
    tapdev_init();
    clock_init();
    timer_set(&periodic_timer, CLOCK_SECOND / 2);
    timer_set(&arp_timer, CLOCK_SECOND * 10);

    // Init uIP
    uip_init();

#ifndef __DHCPC_H__
    // Set the IP address of this host
    uip_ipaddr(ipaddr, HostIpAddress[0], HostIpAddress[1],
                       HostIpAddress[2], HostIpAddress[3]);
    uip_sethostaddr(ipaddr);

    uip_ipaddr(ipaddr, RoutIpAddress[0], RoutIpAddress[1],
                       RoutIpAddress[2], RoutIpAddress[3]);
    uip_setdraddr(ipaddr);

    uip_ipaddr(ipaddr, NetMask[0], NetMask[1], NetMask[2], NetMask[3]);
    uip_setnetmask(ipaddr);
#else
    printf("P: DHCP Supported\n\r");
    uip_ipaddr(ipaddr, 0, 0, 0, 0);
    uip_sethostaddr(ipaddr);
    uip_ipaddr(ipaddr, 0, 0, 0, 0);
    uip_setdraddr(ipaddr);
    uip_ipaddr(ipaddr, 0, 0, 0, 0);
    uip_setnetmask(ipaddr);
#endif

    uip_setethaddr(MacAddress);

    app_init();

    while(1) {
        uip_len = tapdev_read();
        if(uip_len > 0) {
            if(BUF->type == htons(UIP_ETHTYPE_IP)) {
                uip_arp_ipin();
                uip_input();
                /* If the above function invocation resulted in data that
                should be sent out on the network, the global variable
                uip_len is set to a value > 0. */
                if(uip_len > 0) {
                    uip_arp_out();
                    tapdev_send();
                }
            } else if(BUF->type == htons(UIP_ETHTYPE_ARP)) {
                uip_arp_arpin();
                /* If the above function invocation resulted in data that
                should be sent out on the network, the global variable
                uip_len is set to a value > 0. */
                if(uip_len > 0) {
                    tapdev_send();
                }
            }
        } else if(timer_expired(&periodic_timer)) {
            timer_reset(&periodic_timer);
            for(i = 0; i < UIP_CONNS; i++) {
                uip_periodic(i);
                /* If the above function invocation resulted in data that
                   should be sent out on the network, the global variable
                   uip_len is set to a value > 0. */
                if(uip_len > 0) {
                  uip_arp_out();
                  tapdev_send();
                }
            }
#if UIP_UDP
            for(i = 0; i < UIP_UDP_CONNS; i++) {
                uip_udp_periodic(i);
                /* If the above function invocation resulted in data that
                   should be sent out on the network, the global variable
                   uip_len is set to a value > 0. */
                if(uip_len > 0) {
                    uip_arp_out();
                    tapdev_send();
                }
            }
#endif /* UIP_UDP */
      
            /* Call the ARP timer function every 10 seconds. */
            if(timer_expired(&arp_timer)) {
                timer_reset(&arp_timer);
                uip_arp_timer();
            }
        }

        // Display Statistics
        if ( USART_IsDataAvailable((AT91S_USART *)AT91C_BASE_DBGU) ) {

            EmacStats    stats;

            DBGU_GetChar();
            EMAC_GetStatistics(&stats, 1);
            printf("=== EMAC Statistics ===\n\r");
            printf(" .tx_packets = %d\n\r", stats.tx_packets);
            printf(" .tx_comp = %d\n\r", stats.tx_comp);
            printf(" .tx_errors = %d\n\r", stats.tx_errors);
            printf(" .collisions = %d\n\r", stats.collisions);
            printf(" .tx_exausts = %d\n\r", stats.tx_exausts);
            printf(" .tx_underruns = %d\n\r", stats.tx_underruns);
            printf(" .rx_packets = %d\n\r", stats.rx_packets);
            printf(" .rx_eof = %d\n\r", stats.rx_eof);
            printf(" .rx_ovrs = %d\n\r", stats.rx_ovrs);
            printf(" .rx_bnas = %d\n\r", stats.rx_bnas);
        }
    }

    return 0;
}


